
(*
 * 3-D Cutting tools
 *
 * LablGL   version by Jacques Garrigue
 * LablGLUT version by Issac Trotts
 * gears.ml Brain Paul
 * modified by Hiromitsu Morita
 *)

open Printf;;

let pi = 4. *. atan 1.

(*
 * Draw cutting tool. You'll probably want to call this function when
 * building a display list since we do a lot of trig here.
 *
 * Input:    inner_radius - radius of hole at center
 *                 outer_radius - radius at center of teeth
 *                 width - width of gear
 *                 teeth - number of teeth
 *                 tooth_depth - depth of tooth
 *)
let cuttingtool ~inner ~outer ~width ~teeth ~tooth_depth =
  let r0 = inner
  and r1 = outer -. tooth_depth /. 2.0
  and r2 = outer +. tooth_depth /. 2.0 in

  let ta = 2.0 *. pi /. float teeth in
  let da = ta /. 4.0 in

  GlDraw.shade_model `flat;
  
  GlDraw.normal ~z:1.0 ();
  
  let vertex ~r ~z ?(s=0) i =
    let angle = float i *. ta +. float s *. da in
    GlDraw.vertex ~x:(r *. cos angle) ~y:(r *. sin angle) ~z ()
  in

  (* draw front face *)
  let z = width *. 0.5 in
  GlDraw.begins `quad_strip;
  for i=0 to teeth do
    vertex i ~r:r0 ~z;
    vertex i ~r:r1 ~z;
    vertex i ~r:r0 ~z;
    vertex i ~r:r1 ~z ~s:3;
  done;
  GlDraw.ends ();

  (* draw front sides of teeth *)
  GlDraw.begins `quads;
  for i=0 to teeth - 1 do
    if i mod 4 = 0 then
      begin
	vertex i ~r:r1 ~z;
	vertex i ~r:r2 ~s:1 ~z;
	vertex i ~r:r2 ~s:2 ~z;
	vertex i ~r:r1 ~s:3 ~z;
      end
  done;
  GlDraw.ends ()

let workpiece ~x1 ~y1 ~x2 ~y2 ~width =
  GlDraw.shade_model `flat;
  GlDraw.normal ~z:1.0 ();

  (* draw front face *)
  let z = width *. 0.4 in
  GlDraw.begins `polygon;
  GlDraw.vertex ~x:x1 ~y:y1 ~z ();
  GlDraw.vertex ~x:x2 ~y:y1 ~z ();
  GlDraw.vertex ~x:x2 ~y:y2 ~z ();
  GlDraw.vertex ~x:x1 ~y:y2 ~z ();
  GlDraw.ends ();




class view ~cuttingtool1 ~workpiece1 ~workpiece2 ~limit = object (self)
  val mutable view_rotx = 0.0
  val mutable view_roty = 0.0
  val mutable view_rotz = 0.0
  val mutable time = 0
  val mutable center_y = 2.0
  val mutable angle = 0.0
  val mutable count = 1
      
  method draw =
    GlClear.clear [`color;`depth];
    
    GlMat.push ();
    GlMat.rotate ~angle:view_rotx ~x:1.0 ();
    GlMat.rotate ~angle:view_roty ~y:1.0 ();
    GlMat.rotate ~angle:view_rotz ~z:1.0 ();

    GlMat.push ();
    GlMat.translate ~x:(0.) ~y:(0.) ();
    GlMat.rotate ~angle:0. ~z:1.0 ();
    GlList.call workpiece1;
    GlMat.pop ();

    GlMat.push ();
    GlMat.translate ~x:(0.) ~y:(0.) ();
    GlMat.rotate ~angle:0. ~z:1.0 ();
    GlList.call workpiece2;
    GlMat.pop ();
  
    GlMat.push ();
    GlMat.translate ~x:(0.) ~y:(center_y) ();
    GlMat.rotate ~angle:angle ~z:1.0 ();
    GlList.call cuttingtool1;
    GlMat.pop ();

    Glut.swapBuffers();
    
    count <- count + 1;
    if count =limit then exit 0

  method idle =
    time <- time + 1;
    angle <- -2. *. (float time);
    center_y <- 2.0 +. 0.5 *. sin (8. *. pi *. (float time) /. 180.);
    self#draw

  method reshape w h =
    GlDraw.viewport ~x:0 ~y:0 ~w:w ~h:h;
    GlMat.mode `projection;
    GlMat.load_identity ();
    let r = float w /. float h in
    let r' = 1. /. r in
    if (w>h) then
      GlMat.frustum ~x:(-. r,r) ~y:(-1.0,1.0) ~z:(5.0,60.0)
    else
      GlMat.frustum ~x:(-1.0,1.0) ~y:(-.r',r') ~z:(5.0,60.0);
    
    GlMat.mode `modelview;
    GlMat.load_identity();
    GlMat.translate ~z:(-40.0) ();
    GlClear.clear[`color;`depth]
end
    
let init () =
  let pos = 5.0, 5.0, 10.0, 0.0
  and lightglay = 0.8, 0.8, 0.8, 0.8
  and glay = 0.4, 0.4, 0.4, 0.4 in
  
  GlLight.light ~num:0 (`position pos);
  List.iter Gl.enable
    [`cull_face;`lighting;`light0;`depth_test;`normalize];

  let create_workpiece ~x1 ~y1 ~x2 ~y2 ~width ~color =
    let list = GlList.create `compile in
    GlLight.material ~face:`front (`ambient_and_diffuse color);
    workpiece ~x1 ~y1 ~x2 ~y2 ~width;
    GlList.ends ();
    list
  in

  
  let create_cuttingtool ~inner ~outer ~width ~teeth ~color =
    let list = GlList.create `compile in
    GlLight.material ~face:`front (`ambient_and_diffuse color);
    cuttingtool ~inner ~outer ~width ~teeth ~tooth_depth:0.7;
    GlList.ends ();
    list
  in

  let wp1 = create_workpiece ~x1:(-10.) ~y1:(-5.) ~x2:1. ~y2:(-0.5) ~width:3.0 ~color:lightglay
  and wp2 = create_workpiece ~x1:1. ~y1:(-5.) ~x2:10. ~y2:0.5 ~width:3.0 ~color:lightglay
  and ct = create_cuttingtool ~inner:0.0 ~outer:1.4 ~width:3.0 ~teeth:16 ~color:glay in
  (wp1, wp2, ct)

    
let () =
  ignore(Glut.init Sys.argv);
  Glut.initDisplayMode ~alpha:true ~double_buffer:true ~depth:true () ;
  Glut.initWindowSize ~w:300 ~h:300;
  ignore (Glut.createWindow ~title:"cutting tool demo");
  let (wp1, wp2, ct) = init() in
  let view = new view ~workpiece1:wp1 ~workpiece2:wp2 ~cuttingtool1:ct ~limit:0 in
  let _reshape ~w ~h = 
    view#reshape w h 
  in 
  let _keyboard_callback ~key ~x ~y = 
    match (char_of_int key) with 'q' -> exit 0; | _ -> () 
  in
  
  Glut.keyboardFunc ~cb:_keyboard_callback ;
  Glut.reshapeFunc ~cb:_reshape ;
  Glut.displayFunc ~cb:(fun () -> view#draw) ;
  Glut.idleFunc(Some (fun () -> view#idle));
  Glut.mainLoop()
