
(* 横ベクトルの計算 *)
let plus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) +. v.(n))

let minus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) -. v.(n))

let prod_row c v =
  let dim = Array.length v in
  Array.init dim (fun n -> c *. v.(n))

(* 行列の計算 *)
let ( +! ) m1 m2 =
  let dim = Array.length m1 in
  Array.init dim (fun n -> plus_row m1.(n) m2.(n))

let ( -! ) m1 m2 =
  let dim = Array.length m2 in
  Array.init dim (fun n -> minus_row m1.(n) m2.(n))

let ( *< ) c m =
  let dim = Array.length m in
  Array.init dim (fun n -> prod_row c m.(n))

let scalar u v i j =
  let l = Array.length u.(0) in
  let uv = Array.init l (fun n -> u.(i).(n) *. v.(n).(j)) in
  Array.fold_left (+.) 0. uv

(* 行列の積 *)
let ( *! ) a b =
  let row = Array.length a in
  let col = Array.length b.(0) in
  let ab_row k = Array.init col (fun n -> scalar a b k n) in
  Array.init row ab_row

let tr m =
  [|
    [| m.(0).(0); m.(1).(0) |];
    [| m.(0).(1); m.(1).(1) |]
  |]

let inv m =
  let det = m.(0).(0) *. m.(1).(1) -. m.(0).(1) *. m.(1).(0) in
  (1. /. det) *< [| [| m.(1).(1); -. m.(1).(0) |]; [| -. m.(0).(1); m.(0).(0) |] |]

(*
  ns : ステップ数
  y_ob : 観測列
  u_in : 入力列
  mu0 : 初期状態推定値
  sigma0 : 初期誤差共分散行列
  a_mat, b_mat, c_mat, q_mat, r_mat : カルマンフィルタの係数
*)

let lkf ~ns ~y ~u ~mu0 ~sigma0 ~a_mat ~b_mat ~c_mat ~q_mat ~r_mat =
  let rec append_estimate mu_arr sigma n =
    let mu = mu_arr.(n) in
    let mu' = a_mat *! mu +! b_mat *! u.(n) in
    let sigma' = q_mat +! a_mat *! sigma *! (tr a_mat) in
    let yi = y.(n) -! c_mat *! mu' in
    let s_mat = c_mat *! sigma' *! (tr c_mat) +! r_mat in
    let k_mat = sigma' *! (tr c_mat) *! (inv s_mat) in
    let mu_next = mu' +! (k_mat *! yi) in
    let sigma_next = sigma' -! k_mat *! c_mat *! sigma' in
    let mu_arr_next = Array.append mu_arr [| mu_next |] in
    if (n+1) >= ns then mu_arr_next
    else append_estimate mu_arr_next sigma_next (n+1)
  in

  append_estimate [|mu0|] sigma0 0

let pi = 4. *. atan 1.

(* ボックス＝ミュラー法での正規乱数の生成 *)
let nrandom () =
  let r1 = Random.float 1.0 in
  let r2 = Random.float 1.0 in
  let n1 = sqrt (-2. *. log r1) *. sin (2. *. pi *. r2) in
  let n2 = sqrt (-2. *. log r1) *. cos (2. *. pi *. r2) in
  [| [|n1|]; [|n2|] |]


let () =
  let a = [| [| 1.; 0. |]; [| 0.; 1. |] |] in
  let b = [| [| 1.; 0. |]; [| 0.; 1. |] |] in
  let q = [| [| 1.; 0. |]; [| 0.; 1. |] |] in
  let c = [| [| 1.; 0. |]; [| 0.; 1. |] |] in
  let r = [| [| 2.; 0. |]; [| 0.; 2. |] |] in

  let ns = 10 in
  let x0 = [| [|0.|]; [|0.|] |] in

  let rec append_condition x_arr y_arr u_arr n =
    let x_cur = x_arr.(n) in
    let u_cur = u_arr.(n) in
    let v = a *! x_cur in
    let x_next = a *! x_cur +! b *! u_cur +! nrandom () in
    let y_next = c *! x_cur +! nrandom () in
    let x_arr_next = Array.append x_arr [|x_next|] in
    let y_arr_next = Array.append y_arr [|y_next|] in
    let u_arr_next = Array.append u_arr [|u_cur|] in
    if (n+1) >= ns then (x_arr_next, y_arr_next, u_arr_next)
    else append_condition x_arr_next y_arr_next u_arr_next (n+1)
  in

  let (x,y,u) = append_condition [|x0|] [|x0|] [| [| [|2.|]; [|2.|] |] |] 0 in
  let mu0 = [| [|0.|]; [|0.|] |] in
  let sigma0 = [| [|0.;0.|]; [|0.;0.|] |] in

  let x_est = lkf ~ns:ns ~y:y ~u:u ~mu0:mu0 ~sigma0:sigma0 ~a_mat:a ~b_mat:b ~c_mat:c ~q_mat:q ~r_mat:r in

  Array.iter (fun v -> print_endline (string_of_float v.(0).(0))) x_est

