
(*
  OCaml で書いたNewmark β 法による微分方程式の計算
  有限要素法による剛体の回転運動シミュレーション
  lacaml 使用

  ocamlfind ocamlopt -o rotor_dynamics -package lacaml -linkpkg lacaml.cmxa rotor_dynamics.ml
*)

(* finite element method by OCaml *)
(* methode des elements finis par OCaml *)

open Format
open Bigarray

open Lacaml.D
open Lacaml.Io

let pi = 4. *. atan 1.

let mu = 5.0
let l = 28.6
let r = 1.
let omega = 2. *. pi *. 10.
let ip = 0.5
let er = 10.
let ia = 1.0

let m_disk = 10.0
let i_disk = 1.0
let ip_disk = 10.0
let tau = 0.001

let beta = 0.25
let gamma = 0.5

let n_smp = 1000
let dt = 1. /. 500.

(* 横ベクトルの計算 *)
let plus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) +. v.(n))

let minus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) -. v.(n))

let prod_row c v =
  let dim = Array.length v in
  Array.init dim (fun n -> c *. v.(n))

(* 行列の計算 *)
let ( +! ) m1 m2 =
  let dim = Array.length m1 in
  Array.init dim (fun n -> plus_row m1.(n) m2.(n))

let ( -! ) m1 m2 =
  let dim = Array.length m2 in
  Array.init dim (fun n -> minus_row m1.(n) m2.(n))

let ( *< ) c m =
  let dim = Array.length m in
  Array.init dim (fun n -> prod_row c m.(n))

let scalar u v i j =
  let l = Array.length u.(0) in
  let uv = Array.init l (fun n -> u.(i).(n) *. v.(n).(j)) in
  Array.fold_left (+.) 0. uv

(* 行列の積 *)
let ( *! ) a b =
  let row = Array.length a in
  let col = Array.length b.(0) in
  let ab_row k = Array.init col (fun n -> scalar a b k n) in
  Array.init row ab_row

(* クロネッカーのデルタ *)
let kronecker i j = if i = j then 1. else 0.

let e k = Array.init 12 (fun n -> kronecker n k)
let a k = Array.init 8 (fun n -> e (4*k+n))
let tr m =
  let row = Array.length m in
  let col = Array.length m.(0) in
  let tr_row k = Array.init row (fun i -> m.(i).(k)) in
  Array.init col tr_row


(* 質量行列 *)
let mt_00 = 156.
let mt_10 = 0.
let mt_20 = 0.
let mt_30 = -22. *. l
let mt_40 = 54.
let mt_50 = 0.
let mt_60 = 0.
let mt_70 = -13. *. l

let mt_01 = mt_10
let mt_11 = 156.
let mt_21 = -22. *. l
let mt_31 = 0.
let mt_41 = 0.
let mt_51 = 54.
let mt_61 = 13. *. l
let mt_71 = 0.

let mt_02 = mt_20
let mt_12 = mt_21
let mt_22 = 4. *. l *. l
let mt_32 = 0.
let mt_42 = 0.
let mt_52 = -13. *. l
let mt_62 = -3. *. l *. l
let mt_72 = 0.

let mt_03 = mt_30
let mt_13 = mt_31
let mt_23 = mt_32
let mt_33 = 4. *. l *. l
let mt_43 = 13. *. l
let mt_53 = 0.
let mt_63 = 0.
let mt_73 = -3. *. l *. l

let mt_04 = mt_40
let mt_14 = mt_41
let mt_24 = mt_42
let mt_34 = mt_43
let mt_44 = 156.
let mt_54 = 0.
let mt_64 = 0.
let mt_74 = -22. *. l

let mt_05 = mt_50
let mt_15 = mt_51
let mt_25 = mt_52
let mt_35 = mt_53
let mt_45 = mt_54
let mt_55 = 156.
let mt_65 = 22. *. l
let mt_75 = 0.

let mt_06 = mt_60
let mt_16 = mt_61
let mt_26 = mt_62
let mt_36 = mt_63
let mt_46 = mt_64
let mt_56 = mt_65
let mt_66 = 4. *. l *. l
let mt_76 = 0.

let mt_07 = mt_70
let mt_17 = mt_71
let mt_27 = mt_72
let mt_37 = mt_73
let mt_47 = mt_74
let mt_57 = mt_75
let mt_67 = mt_76
let mt_77 = 4. *. l *. l


let mt =
  (mu *. l /. 420. ) *< 
  [|
    [| mt_00; mt_01; mt_02; mt_03; mt_04; mt_05; mt_06; mt_07 |];
    [| mt_10; mt_11; mt_12; mt_13; mt_14; mt_15; mt_16; mt_17 |];
    [| mt_20; mt_21; mt_22; mt_23; mt_24; mt_25; mt_26; mt_27 |];
    [| mt_30; mt_31; mt_32; mt_33; mt_34; mt_35; mt_36; mt_37 |];
    [| mt_40; mt_41; mt_42; mt_43; mt_44; mt_45; mt_46; mt_47 |];
    [| mt_50; mt_51; mt_52; mt_53; mt_54; mt_55; mt_56; mt_57 |];
    [| mt_60; mt_61; mt_62; mt_63; mt_64; mt_65; mt_66; mt_67 |];
    [| mt_70; mt_71; mt_72; mt_73; mt_74; mt_75; mt_76; mt_77 |];
  |]


let mr_00 = 36.
let mr_10 = 0.
let mr_20 = 0.
let mr_30 = 3. *. l
let mr_40 = -36.
let mr_50 = 0.
let mr_60 = 0.
let mr_70 = 3. *. l

let mr_01 = mr_10
let mr_11 = 36.
let mr_21 = -3. *. l
let mr_31 = 0.
let mr_41 = 0.
let mr_51 = -36.
let mr_61 = -3. *. l
let mr_71 = 0.

let mr_02 = mr_20
let mr_12 = mr_21
let mr_22 = 4. *. l *. l
let mr_32 = 0.
let mr_42 = 0.
let mr_52 = 3. *. l
let mr_62 = -. l *. l
let mr_72 = 0.

let mr_03 = mr_30
let mr_13 = mr_31
let mr_23 = mr_32
let mr_33 = 4. *. l *. l
let mr_43 = -3. *. l
let mr_53 = 0.
let mr_63 = 0.
let mr_73 = -. l *. l

let mr_04 = mr_40
let mr_14 = mr_41
let mr_24 = mr_42
let mr_34 = mr_43
let mr_44 = 36.
let mr_54 = 0.
let mr_64 = 0.
let mr_74 = -3. *. l

let mr_05 = mr_50
let mr_15 = mr_51
let mr_25 = mr_52
let mr_35 = mr_53
let mr_45 = mr_54
let mr_55 = 36.
let mr_65 = 3. *. l
let mr_75 = 0.

let mr_06 = mr_60
let mr_16 = mr_61
let mr_26 = mr_62
let mr_36 = mr_63
let mr_46 = mr_64
let mr_56 = mr_65
let mr_66 = 4. *. l *. l
let mr_76 = 0.

let mr_07 = mr_70
let mr_17 = mr_71
let mr_27 = mr_72
let mr_37 = mr_73
let mr_47 = mr_74
let mr_57 = mr_75
let mr_67 = mr_76
let mr_77 = 4. *. l *. l

let mr =
  (mu *. r *. r /. (120. *. l)) *< 
  [|
    [| mr_00; mr_01; mr_02; mr_03; mr_04; mr_05; mr_06; mr_07 |];
    [| mr_10; mr_11; mr_12; mr_13; mr_14; mr_15; mr_16; mr_17 |];
    [| mr_20; mr_21; mr_22; mr_23; mr_24; mr_25; mr_26; mr_27 |];
    [| mr_30; mr_31; mr_32; mr_33; mr_34; mr_35; mr_36; mr_37 |];
    [| mr_40; mr_41; mr_42; mr_43; mr_44; mr_45; mr_46; mr_47 |];
    [| mr_50; mr_51; mr_52; mr_53; mr_54; mr_55; mr_56; mr_57 |];
    [| mr_60; mr_61; mr_62; mr_63; mr_64; mr_65; mr_66; mr_67 |];
    [| mr_70; mr_71; mr_72; mr_73; mr_74; mr_75; mr_76; mr_77 |];
  |]

let g_00 = 0.
let g_10 = -36.
let g_20 = 3. *. l
let g_30 = 0.
let g_40 = 0.
let g_50 = 36.
let g_60 = 3. *. l
let g_70 = 0.

let g_01 = -. g_10
let g_11 = 0.
let g_21 = 0.
let g_31 = 3. *. l
let g_41 = -36.
let g_51 = 0.
let g_61 = 0.
let g_71 = 3. *. l

let g_02 = -. g_20
let g_12 = -. g_21
let g_22 = 0.
let g_32 = -. 4. *. l *. l
let g_42 = 3. *. l
let g_52 = 0.
let g_62 = 0.
let g_72 = l *. l

let g_03 = -. g_30
let g_13 = -. g_31
let g_23 = -. g_32
let g_33 = 0.
let g_43 = 0.
let g_53 = 3. *. l
let g_63 = -. l *. l
let g_73 = 0.

let g_04 = -. g_40
let g_14 = -. g_41
let g_24 = -. g_42
let g_34 = -. g_43
let g_44 = 0.
let g_54 = -36.
let g_64 = -3. *. l
let g_74 = 0.

let g_05 = -. g_50
let g_15 = -. g_51
let g_25 = -. g_52
let g_35 = -. g_53
let g_45 = -. g_54
let g_55 = 0.
let g_65 = 0.
let g_75 = -3. *. l

let g_06 = -. g_60
let g_16 = -. g_61
let g_26 = -. g_62
let g_36 = -. g_63
let g_46 = -. g_64
let g_56 = -. g_65
let g_66 = 0.
let g_76 = -. 4. *. l *. l

let g_07 = -. g_70
let g_17 = -. g_71
let g_27 = -. g_72
let g_37 = -. g_73
let g_47 = -. g_74
let g_57 = -. g_75
let g_67 = -. g_76
let g_77 = 0.

let g =
  (omega *. ip /. (30. *. l)) *<
  [|
    [| g_00; g_01; g_02; g_03; g_04; g_05; g_06; g_07 |];
    [| g_10; g_11; g_12; g_13; g_14; g_15; g_16; g_17 |];
    [| g_20; g_21; g_22; g_23; g_24; g_25; g_26; g_27 |];
    [| g_30; g_31; g_32; g_33; g_34; g_35; g_36; g_37 |];
    [| g_40; g_41; g_42; g_43; g_44; g_45; g_46; g_47 |];
    [| g_50; g_51; g_52; g_53; g_54; g_55; g_56; g_57 |];
    [| g_60; g_61; g_62; g_63; g_64; g_65; g_66; g_67 |];
    [| g_70; g_71; g_72; g_73; g_74; g_75; g_76; g_77 |]
  |]


let k_00 = 12.
let k_01 = 0.
let k_02 = 0.
let k_03 = 6. *. l
let k_04 = -12.
let k_05 = 0.
let k_06 = 0.
let k_07 = 6. *. l

let k_10 = k_01
let k_11 = 12.
let k_12 = -6. *. l
let k_13 = 0.
let k_14 = 0.
let k_15 = -12.
let k_16 = -6. *. l
let k_17 = 0.

let k_20 = k_02
let k_21 = k_12
let k_22 = 4. *. l *. l
let k_23 = 0.
let k_24 = 0.
let k_25 = 6. *. l
let k_26 = 2. *. l *. l
let k_27 = 0.

let k_30 = k_03
let k_31 = k_13
let k_32 = k_23
let k_33 = 4. *. l *. l
let k_34 = -6. *. l
let k_35 = 0.
let k_36 = 0.
let k_37 = 2. *. l *. l

let k_40 = k_04
let k_41 = k_14
let k_42 = k_24
let k_43 = k_34
let k_44 = 12.
let k_45 = 0.
let k_46 = 0.
let k_47 = -6. *. l

let k_50 = k_05
let k_51 = k_15
let k_52 = k_25
let k_53 = k_35
let k_54 = k_45
let k_55 = 12.
let k_56 = 6. *. l
let k_57 = 0.

let k_60 = k_06
let k_61 = k_16
let k_62 = k_26
let k_63 = k_36
let k_64 = k_46
let k_65 = k_56
let k_66 = 4. *. l *. l
let k_67 = 0.

let k_70 = k_07
let k_71 = k_17
let k_72 = k_27
let k_73 = k_37
let k_74 = k_47
let k_75 = k_57
let k_76 = k_67
let k_77 = 4. *. l *. l

let k =
  (er *. ia /. (l *. l *. l)) *<
  [|
    [| k_00; k_01; k_02; k_03; k_04; k_05; k_06; k_07 |];
    [| k_10; k_11; k_12; k_13; k_14; k_15; k_16; k_17 |];
    [| k_20; k_21; k_22; k_23; k_24; k_25; k_26; k_27 |];
    [| k_30; k_31; k_32; k_33; k_34; k_35; k_36; k_37 |];
    [| k_40; k_41; k_42; k_43; k_44; k_45; k_46; k_47 |];
    [| k_50; k_51; k_52; k_53; k_54; k_55; k_56; k_57 |];
    [| k_60; k_61; k_62; k_63; k_64; k_65; k_66; k_67 |];
    [| k_70; k_71; k_72; k_73; k_74; k_75; k_76; k_77 |]
  |]


let a0 = a 0
let tr_a0 = tr a0
let a1 = a 1
let tr_a1 = tr a1


let md_elem i j =
  if (i = 4 && j = 4) || (i = 5 && j = 5) then m_disk
  else if (i = 6 && j = 6) || (i = 7 && j = 7) then i_disk
  else 0.

let md_row i = Array.init 12 (md_elem i)
let md = Array.init 12 md_row

let nd_elem i j =
  if (i = 6 && j = 7) then -. i_disk
  else if (i = 7 && j = 6) then i_disk
  else 0.

let nd_row i = Array.init 12 (nd_elem i)
let nd = Array.init 12 nd_row

let m_sys = (tr_a0 *! (mt +! mr) *! a0) +! (tr_a1 *! (mt +! mr) *! a1) +! md
let g_sys = (tr_a0 *! g *! a0) +! (tr_a1 *! g *! a1) +! nd
let k_sys = (tr_a0 *! k *! a0) +! (tr_a1 *! k *! a1) 
let c_sys = (1.0 *< m_sys) +! (0.5 *< k_sys)




(* 連立方程式 : ax = b の解を計算する関数 *)
let solve a b =
  let a_mat = Mat.of_array a in
  let b_mat = Mat.of_array b in
  gesv a_mat b_mat;
  Mat.to_array b_mat



let newmark_beta n_cur (m_ar2, c_ar2, k_ar2) fn (x_p, v_p, a_p) =
  let v_tmp = v_p +! (dt *. (1. -. gamma)) *< a_p in
  let x_tmp = x_p +! (dt *< v_p) +! (0.5 *. dt *. dt *. (1. -. 2.*.beta)) *< a_p in
  let b_tmp = fn -! (c_ar2 *! v_tmp) -! (k_ar2 *! x_tmp) in
  let a_m = m_ar2 +! ((dt *. gamma) *< c_ar2) +! ((dt *. dt *. beta) *< k_ar2) in
  let a_c = solve a_m b_tmp in
  let v_c = v_p +! (dt *. (1. -. gamma)) *< a_p +! (gamma *< a_c) in
  let x_c = x_p +! (dt *< v_p) +! (0.5 *. dt *. dt *. (1. -. 2.*.beta)) *< a_p +! (beta *. dt *. dt) *< a_c in
  (x_c, v_c, a_c)

let rec dump_newmark_beta n_cur m_ar2 c_ar2 k_ar2 f x_p v_p a_p =
  let t = (float_of_int n_cur) /. (float_of_int n_smp) in
  let fn = f t in
  let (x_c, v_c, a_c) = newmark_beta n_cur (m_ar2, c_ar2, k_ar2) fn (x_p, v_p, a_p) in
  print_string ((string_of_int n_cur) ^ "\t");
  Array.iter (fun arr -> print_string ((string_of_float arr.(0)) ^ "\t")) x_c;
  print_endline "";
  if n_cur + 1 >= n_smp then () else dump_newmark_beta (n_cur+1) m_ar2 c_ar2 k_ar2 f x_c v_c a_c


let () =
  let x0 = Array.make 12 [|0.|] in
  let v0 = Array.make 12 [|0.|] in
  let r = 5e-6 in
  let f_ext t =
    [|
      [|0.|];
      [|0.|];
      [|0.|];
      [|0.|];
      [| m_disk *. r *. omega *. omega *. (cos (omega *. t)) |];
      [| m_disk *. r *. omega *. omega *. (sin (omega *. t)) |];
      [| -.(i_disk -. ip_disk) *. tau *. omega *. omega *. (sin (omega *. t)) |];
      [| (ip_disk -. i_disk) *. tau *. omega *. omega *. (cos (omega *. t))|];
      [|0.|];
      [|0.|];
      [|0.|];
      [|0.|];
    |]
  in

  let f0 = f_ext 0. in
  let b0 = f0 -! ((c_sys +! g_sys) *! v0) -! (k_sys *! x0) in
  let a0 = solve m_sys b0 in
  dump_newmark_beta 1 m_sys (c_sys +! g_sys) k_sys f_ext x0 v0 a0
