
(* 遅れを含む微分方程式 *)

let m = 1.
let c = 1.
let k = 100.
let del = 100.

let f t x y =
  [|
    [| x.(1).(0) |];
    [| -. (c /. m) *. x.(1).(0) -. ((k /. m) +. del) *. x.(0).(0) -. del *. y.(0).(0) |]
  |]

let phi t =
  [|
    [| 1. |];
    [| 0. |]
  |]

(* 横ベクトルの計算 *)
let plus_row u v =
  Array.init (Array.length u) (fun n -> u.(n) +. v.(n))
let minus_row u v =
  Array.init (Array.length u) (fun n -> u.(n) -. v.(n))
let ctimes_row c v =
  Array.init (Array.length v) (fun n -> c *. v.(n))

(* 行列の計算 *)
let ( +! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> plus_row m1.(n) m2.(n))
let ( -! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> minus_row m1.(n) m2.(n))
let ( *! ) c m =
  Array.init (Array.length m) (fun n -> ctimes_row c m.(n))

(* ルンゲ=クッタ法で次のサンプル点を計算 *)
let runge_kutta_core f x0 ys1 ys2 ys3 ys4 h n =
  let t = (float_of_int n) *. h in
  let k1 = f t x0 ys1.(n) in
  let x1 = x0 +! ((0.5 *. h) *! k1) in
  let k2 = f (t +. 0.5 *. h) x1 ys2.(n) in
  let x2 = x0 +! ((0.5 *. h) *! k2) in
  let k3 = f (t +. 0.5 *. h) x2 ys3.(n) in
  let x3 = x0 +! (h *! k3) in
  let k4 = f (t +. h) x3 ys4.(n) in
  let x4 = x0 +! ((h /. 6.) *! (k1 +! (2. *! k2) +! (2. *! k3) +! k4)) in
  (x1, x2, x3, x4)

let solve ~f ~x0 ~ys1 ~ys2 ~ys3 ~ys4 ~h =
  let n_len = Array.length ys1 in
  let (x1_1, x1_2, x1_3, x1) = runge_kutta_core f x0 ys1 ys2 ys3 ys4 h 0 in
  let rec append_solution xs1 xs2 xs3 xs4 xp n =
    let (xn_1, xn_2, xn_3, xn) = runge_kutta_core f xp ys1 ys2 ys3 ys4 h n in 
    let xs1_n = Array.append xs1 [| xn_1 |] in
    let xs2_n = Array.append xs1 [| xn_2 |] in
    let xs3_n = Array.append xs1 [| xn_3 |] in
    let xs4_n = Array.append xs1 [| xn |] in
    if (n+1) >= n_len then (xs1_n, xs2_n, xs3_n, xs4_n)
    else append_solution xs1_n xs2_n xs3_n xs4_n xn (n+1)
  in

  append_solution [|x1_1|] [|x1_2|] [|x1_3|] [|x1|] x1 1



let print_elem x_elem =
  print_endline (string_of_float x_elem.(0).(0))


let () =
  let ys1 = Array.make 500 [| [|0.|];[|0.|] |] in
  let ys2 = Array.make 500 [| [|0.|];[|0.|] |] in
  let ys3 = Array.make 500 [| [|0.|];[|0.|] |] in
  let ys4 = Array.make 500 [| [|0.|];[|0.|] |] in

  let x0 = [| [|1.|];[|0.|] |] in
  let (ys1_1, ys2_1, ys3_1, xs1) = solve ~f:f ~x0:x0 ~ys1:ys1 ~ys2:ys2 ~ys3:ys3 ~ys4:ys4 ~h:0.01 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) xs1;
  let (ys1_2, ys2_2, ys3_2, xs2) = solve ~f:f ~x0:xs1.(499) ~ys1:ys1_1 ~ys2:ys2_1 ~ys3:ys3_1 ~ys4:xs1 ~h:0.01 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) xs2;
  let (ys1_3, ys2_3, ys3_3, xs3) = solve ~f:f ~x0:xs2.(499) ~ys1:ys1_2 ~ys2:ys2_2 ~ys3:ys3_2 ~ys4:xs2 ~h:0.01 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) xs3;
  let (ys1_4, ys2_4, ys3_4, xs4) = solve ~f:f ~x0:xs3.(499) ~ys1:ys1_3 ~ys2:ys2_3 ~ys3:ys3_3 ~ys4:xs3 ~h:0.01 in
  Array.iter (fun elm -> print_endline (string_of_float elm.(0).(0))) xs4


